// unraid.c - Copyright 2006, Brantley Coile Company
#define _FILE_OFFSET_BITS 64 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdint.h>
#include <assert.h>
#include <errno.h>

#define nil NULL
#define BADSHELF 0x10000	// illegal AoE shelf address
#define CHUNKSECTORS 128LL	// 64 KB chunk size
#define RDATASTART 3LL		// raid data sector offset in raiddev
#define MIN(a, b) ((a) > (b) ? (b) : (a))
#define CORAID_MAGIC "coraid00"
#define SRCAT_VERSION "alpha-0"

typedef uint8_t uchar;
typedef int64_t vlong;

enum { 
	MAXF	= 30,
};

const char *usage = "srcat -V  # (print version)\n"
	"srcat -s {shelfno} [options] file1 [file2 ...]\n"
	"options:\n"
	"\t-o {sector offset}\n"
	"\t-o {KB offset}K\n"
	"\t-l {sector length}\n"
	"\t-l {KB length}K\n"
;

typedef struct Spare Spare;
typedef struct Mask Mask;
typedef struct Device Device;
typedef struct Online Online;
typedef struct Super Super;

struct Spare
{
	Spare	*next;
	char		*dev;
};

struct Mask
{
	Mask	*next;
	int		slot;
	ulong	ts;
	char		*mask;
};

struct Device
{
	Device	*next;
	Device	*prev;
	char	*dev;
	char	*raidtype;
	int	ndevs;
	int	l, p, d;		// logical, part, device
	ulong	ts;			// timestamp
	char		*state;
	int		nrow;		// parity build row
};

struct Online
{
	Online	*next;
	int		slot;
};

struct Super 
{
	char 		magic[8];
	uchar 	length[2];
	uchar 	sectoralign[512 - 8 - 2];
	char 		conf[1024];
};

int		dbflag;		// force
Spare	*spares;
Device	*devices;
Mask	*masks;
Online	*onlines;

int rdopen(const char *path)
{
	int fd = open(path, O_RDONLY);
	
	if (fd < 0) {
		fprintf(stderr, 
			"srcat Error: could not open %s for reading: %s\n",
			path, strerror(errno));
		exit(EXIT_FAILURE);
	}
	return fd;
}

off_t
xlseek(const char *name, int fd, off_t offset, int whence)
{
	off_t res = lseek(fd, offset, whence);

	if (res < 0) {
		fprintf(stderr, 
			"srcat Error: could not seek on %s: %s\n",
			name, strerror(errno));
		exit(EXIT_FAILURE);
	}
	return res;
}	

ssize_t
xread(const char *name, int fd, void *buf, size_t count)
{
	ssize_t n = read(fd, buf, count);

	if (n <= 0) {
		fprintf(stderr,
			"srcat Error: cannot read %u bytes of %s: %s\n",
			count, name, strerror(errno));
		exit(EXIT_FAILURE);
	}
	return n;
}

void *
xmallocz(size_t n, int clear)
{
	void *p;

	if (clear)
		p = calloc(1, n);
	else
		p = malloc(n);

	if (!p) {
		perror("mem alloc");
		exit(EXIT_FAILURE);
	}

	return p;
}

char *
xstrdup(char *p)
{
	p = strdup(p);

	if (!p) {
		perror("strdup");
		exit(EXIT_FAILURE);
	}

	return p;
}

int
tokenize(char *str, char **args, int maxargs)
{
	char *p;
	int i;

	for (i = 0; i < maxargs; str = NULL) {
		p = strtok(str, "\t\r\n ");
		if (p)
			args[i++] = xstrdup(p);
		else
			break;
	}
	
	return i;
}

void
noteonline(int slot)
{
	Online *p;

	p = xmallocz(sizeof *p, 1);
	p->slot = slot;
	p->next = onlines;
	onlines = p;
}

void
emitonlines(void)
{
	Online *p;
	
	for (p = onlines; p; p = p->next)
		fprintf(stderr, "online %d\n", p->slot);
}

int
getcfstr(char *name, char **f)
{
	char buf[4096];
	int fd, n;
	Super *p;
	char *cfg;
	unsigned int cfglen = 0;

	fd = rdopen(name);
	n = xread(name, fd, buf, sizeof buf);
	p = (Super *) buf;
	assert(!memcmp(p->magic, CORAID_MAGIC, sizeof p->magic));
	cfglen = p->length[1] | p->length[0] << 8;
	assert(cfglen <= sizeof(p->conf));
	cfg = xmallocz(cfglen+1, 1);
	memmove(cfg, p->conf, cfglen);
	f[0] = xstrdup(name);
	n = tokenize(cfg, f+1, MAXF);
	close(fd);
	return n+1;
}

void
addspare(char **f)	// add new spare device
{
	Spare *p;

	p = xmallocz(sizeof *p, 1);
	p->next = spares;
	spares = p;
	p->dev = xstrdup(f[0]);
}

void
emitspares(void)	// write out the spare commands
{
	Spare *p;

	for (p = spares; p; p = p->next)
		fprintf(stderr, "spare %s\n", p->dev);
}

void
setlpd(Device *p, char *lpd)
{
	char *cp;

	cp = strchr(lpd, '.');
	if (cp == nil)
		return;
	*cp++ = 0;
	p->l = atoi(lpd);
	lpd = cp;
	cp = strchr(lpd, '.');
	if (cp == nil)
		return;
	*cp++ = 0;
	p->p = atoi(lpd);
	p->d = atoi(cp);
}

void
addmasks(int nf, char **f, int slot)	// add masks to the list
{
	Mask *p, **q, *np;
	char *mask;

next:
	while (nf-- > 0) {
		mask = *f++;
		for (p = masks; p; p = p->next)
			if (p->slot == slot)
			if (strcmp(p->mask, mask) == 0)
				goto next;
		np = xmallocz(sizeof *np, 1);
		np->slot = slot;
		np->mask = xstrdup(mask);
		q = &masks;
		for (p = masks; p; p = p->next) {
			if (slot == p->slot)
				break;
			q = &p->next;
		}
		np->next = *q;
		*q = np;
	}
}

void
rmmasks(int slot, ulong ts)	// kill masks for this slot
{
	Mask *p, **q;

	q = &masks;
	for (p = masks; p; p = p->next) {
		if (p->slot == slot && p->ts < ts) {
			*q = p->next;	// don't bother freeing struct
			continue;
		}
		q = &p->next;
	}
}

void
emitmasks(void)	// output the mask commands
{
	Mask *p;
	int oldslot;

	p = masks;
	if (p == nil)
		return;
	oldslot = p->slot;
	while (p) {
		fprintf(stderr, "mask %d +%s", p->slot, p->mask);
		for (;;) {
			p = p->next;
			if (p == nil) {
				putc('\n', stderr);
				break;
			}
			if (p->slot != oldslot) {
				oldslot = p->slot;
				putc('\n', stderr);
				break;
			}
			fprintf(stderr, " +%s", p->mask);
		}
	}
}

Device *
newdev(char **f)
{
	Device *p;

	p = xmallocz(sizeof *p, 1);
	p->dev = xstrdup(f[0]);
	p->raidtype = xstrdup(f[4]);
	p->ndevs = atoi(f[3]);
	setlpd(p, f[2]);
	p->ts = atol(f[7]);
	p->state = xstrdup(f[5]);
	p->nrow = atoi(f[6]);
	return p;
}

void
printdev(Device *p)
{
	fprintf(stderr, " dev %6s,", p->dev);
	fprintf(stderr, " raidtype %6s,", p->raidtype);
	fprintf(stderr, " ndevs %3d,", p->ndevs);
	fprintf(stderr, "%2d %2d %2d ,", p->l, p->p, p->d);
	fprintf(stderr, " ts %lu,", p->ts);
//	fprintf(stderr, " nrow %8d,", p->nrow);
	fprintf(stderr, " state %s\n", p->state);
}

void
printdevices(void)
{
	Device *p;

	for (p = devices; p; p = p->next)
		printdev(p);
}

int
cmpare(Device *q, Device *p)	// <0 if less, ==0 if eq, >0 if greater
{
	if (q->l < p->l)
		return -1;
	if (q->l > p->l)
		return 1;
	if (q->p < p->p)
		return -1;
	if (q->p > p->p)
		return 1;
	if (q->d < p->d)
		return -1;
	if (q->d > p->d)
		return 1;
	return 0;
}

Device *
finddev(Device *q)	// lookup in our list of devices
{
	Device *p;

	for (p = devices; p; p = p->next)
		if (cmpare(q, p) == 0)
			return p;
	return nil;
}

void
replace(Device *op, Device *np)	// replace old with new
{
	np->next = op->next;
	np->prev = op->prev;
	if (np->next)
		np->next->prev = np;
	if (np->prev)
		np->prev->next = np;
	else
		devices = np;
	free(op);
}

void
insert(Device *p)	// put this device in the list
{
	Device *q;

	q = devices;
	if (q == nil) {
		devices = p;
		return;
	}
	while (cmpare(p, q) >= 0 && q->next != nil)
		q = q->next;
	if (q->next == nil && cmpare(p, q) >= 0) {
		q->next = p;
		p->prev = q;
		return;
	}
	p->next = q;
	p->prev = q->prev;
	q->prev = p;
	if (p->prev == nil)
		devices = p;
	else
		p->prev->next = p;
}

void
adddevice(int nf, char **f)
{
	Device *dp, *ndp;

	ndp = newdev(f);
	dp = finddev(ndp);
	if (dp) {
		if (ndp->ts > dp->ts) {
			rmmasks(dp->l, dp->ts);
			replace(dp, ndp);
		} else {
			free(ndp);
			return;
		}
	} else
		insert(ndp);
	if (nf > 8) 
		addmasks(nf - 8, &f[8], ndp->l);
}


Device *
appendfake(Device *p, int l, int part, int d, char *raidtype)
{
	Device *np;

	np = xmallocz(sizeof *np, 1);
	np->dev = "fake";
	np->raidtype = xstrdup(raidtype);
	np->l = l;
	np->p = part;
	np->d = d;
	np->ndevs = p->ndevs;
	np->state = "normal";
	np->next = p->next;
	np->prev = p;
	p->next = np;
	if (np->next)
		np->next->prev = np;
	return np;
}

// remove span of devices inclusive
void
rmspan(Device *start, Device *p)	// careful not to bother p's next pointer
{
	if (start->prev)
		start->prev->next = p->next;
	else
		devices = p->next;
	if (p->next)
		p->next->prev = start->prev;
}

Device *
killlblade(int slot)	// kill all the nodes with slot; will be contiguious
{
	Device *p, *q;

	p = devices;
	while (p && p->l != slot)
		p = p->next;
	if (p == nil)
		return nil;
	q = p;
	while (q->next && q->next->l == slot)
		q = q->next;
	rmspan(p, q);
	rmmasks(slot, ~0);
	return q->next;
}

void
insertfake(Device *p, int slot, int part, int dev, char *raidtype)
{
	Device *np;

	np = xmallocz(sizeof *np, 1);
	np->dev = "fake";
	np->raidtype = xstrdup(raidtype);
	np->l = slot;
	np->p = part;
	np->d = dev;
	np->ndevs = p->ndevs;
	np->state = "normal";
	np->next = p;
	np->prev = p->prev;
	p->prev = np;
	if (np->prev)
		np->prev->next = np;
	else
		devices = np;
}

/* go through list ditching devices in parts with inconsistent timestamps. */
void
tscheck(void)
{
	Device *p, *np;

	if (devices == nil)
		return;
	p = devices;
loop:
	np = p->next;
	if (np == nil)
		return;
	if (p->l == np->l && p->p == np->p) {
		if (p->ts > np->ts) {
			rmspan(np, np);
			goto loop;
		}
		if (p->ts < np->ts)
			rmspan(p, p);
	}
	p = np;
	goto loop;
}

void
mkfakes(void)		//  insert fakes where required
{
	Device *p, *q;
	char *rtype;
	int ol, op, od, ndevs, i;

	p = devices;
A:
	if (p == nil)
		return;
	ol = p->l;
	if (p->p != 0) {
		p = killlblade(ol);
		goto A;
	}
B:
	op = p->p;
	ndevs = p->ndevs;
	rtype = p->raidtype;
	for (i = 0; i < p->d; i++)
		insertfake(p, ol, op, i, rtype);
	for (;;) {
		od = p->d;
		q = p;
		p = p->next;
		if (p == nil || p->p != op || p->l != ol)
			for (i = od+1; i < ndevs; i++)
				q = appendfake(q, ol, op, i, rtype);
		if (p == nil)
			return;
		if (p->l != ol)
			goto A;
		if (p->d >= ndevs) {
			rmspan(p, p);
			continue;
		}
		if (p->p == op+1)
			goto B;
		if (p->p != op) {
			p = killlblade(ol);
			goto A;
		}
		for (i = od+1; i < p->d; i++)
			insertfake(p, ol, op, i, rtype);
	}
}

char *
suffix(Device *p)
{
	if (strcmp(p->state, "clean") == 0)
		return ":c";
	else if (strcmp(p->state, "failed") == 0)
		return ":f";
	else if (strcmp(p->state, "replaced") == 0)
		return ":r";
	else
		return "";
}

void
emitmake(void)	// writeout make and grow commands
{
	Device *p;
	int l, part;

	p = devices;
	if (p == nil)
		return;
	while (p) {
		fprintf(stderr, 
			"make -r %d %s %s%s ", 
			p->l, p->raidtype, p->dev, suffix(p));
		noteonline(p->l);
		l = p->l;
		do {
			part = p->p;
			for (p = p->next; p && p->p == part && p->l == l; p = p->next)
				fprintf(stderr,
					 "%s%s ",
					p->dev, suffix(p));
			putc('\n', stderr);
			if (p && p->l == l)
				fprintf(stderr, 
					"grow %d %s %s%s ", 
					p->l, p->raidtype, p->dev, suffix(p));
		} while (p && p->l == l);
	}
}		

void
emit(void)		// write out all the commands
{
	emitspares();
	emitmake();
	emitmasks();
	emitonlines();
}

vlong
strsectors(const char *str)
{
	char *end;
	vlong n;

	n = strtoll(str, &end, 0);
	if (end == str || errno == ERANGE) {
		fputs("bad sector count\n", stderr);
		fputs(usage, stderr);
		exit(EXIT_FAILURE);
	}
	if (*end == 'k' || *end == 'K')
		n /= 2;

	return n;
}

vlong raidsectors(Device *devs)
{
	Device *p = devs;
	vlong n;
	vlong min = 0;
	int i, end, fd;

	for (i = 0, end = p->ndevs; p && i < end; p = p->next, ++i) {
		assert(strcmp(p->dev, "fake"));	// todo: handle fakes
		fd = rdopen(p->dev);
		n = xlseek(p->dev, fd, 0, SEEK_END);
		close(fd);
		if (min == 0 || min < n)
			min = n;
	}
	min /= 512;
	min -= 3;	// three sector disk header
	if (!strcmp(devs->raidtype, "raid5"))
		return min * (i-1);
	else if (!strcmp(devs->raidtype, "raid10")
		 || !strcmp(devs->raidtype, "raid1"))
		return min * (i/2);
	else
		return min * i;
}

// based on HowRaid
void
putraid5(Device *devs, vlong off, int n)
{
	static char buf[1024 * 64];
	vlong vbn = off / CHUNKSECTORS;
	vlong pbn = vbn / (devs->ndevs - 1);
	vlong disk = vbn % devs->ndevs;
	int i, fd;
	int chunkoff, chunkn;
	Device *p;

	assert(n <= CHUNKSECTORS);
	if (n == 0)
		return;
	for (i = 0, p = devs; p && i < disk; ++i, p = p->next)
		;
	assert(p && i == disk);
	
	fd = rdopen(p->dev);
	xlseek(p->dev, fd, 
	       (pbn * CHUNKSECTORS * 512) + (RDATASTART * 512),
	       SEEK_SET);
	xread(p->dev, fd, buf, CHUNKSECTORS * 512);
	close(fd);

	chunkoff = off % CHUNKSECTORS;
	chunkn = CHUNKSECTORS - chunkoff;
	fwrite(&buf[chunkoff * 512], 512, MIN(n, chunkn), stdout);

	if (n - chunkn > 0)
		putraid5(devs, off + chunkn, n - chunkn);
}

void
putdata(vlong start, vlong len)
{
	Device *p = devices;
	vlong i;
	vlong end;

	if (!p || len == 0)
		return;

	if (!strcmp("raid5", p->raidtype)) {
		end = raidsectors(p);
		if (len != -1)
			end = len;
		for (i = start; end - i > CHUNKSECTORS; i += CHUNKSECTORS)
			putraid5(p, i, CHUNKSECTORS);
		putraid5(p, i, end - i);
	} else {
		fprintf(stderr,
			"unsupported raid type: %s\n",
			p->raidtype);
	}
}

int 
main(int argc, char *argv[])
{
	int nf, i;
	int sh = BADSHELF;
	char *f[MAXF];
	int c;
	vlong start = 0;
	vlong len = -1;

	while ((c = getopt(argc, argv, "Vs:o:l:")) != -1) {
		switch (c) {
		case 'V':
			puts(SRCAT_VERSION);
			exit(EXIT_SUCCESS);
			break;
		case 's':	// dirname
			sh = atoi(optarg);
			break;
		case 'd':
			dbflag = 1;
			break;
		case 'o':
			start = strsectors(optarg);
			break;
		case 'l':
			len = strsectors(optarg);
			break;
		case '?':
			fputs(usage, stderr);
			exit(EXIT_FAILURE);
			break;
		}
	}
	if (sh == BADSHELF) {
		fputs(usage, stderr);
		exit(EXIT_FAILURE);
	}

	fputs("Reading config information from drives ... ", stderr);
	for (i = optind; i < argc; ++i) {
		nf = getcfstr(argv[i], f);
		if (nf < 3)
			continue;
		if (atoi(f[1]) != sh)
			continue;
		if (strcmp(f[2], "spare") == 0)
			addspare(f);
		else
			adddevice(nf, f);
	}
	fputs("done.\n", stderr);
	if (dbflag)
		printdevices();
	tscheck();
	mkfakes();
	emit();
	putdata(start, len);

	return 0;
}
